@@ -5,7 +5,7 @@
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2012	PgPool Global Development Group
+ * Copyright (c) 2003-2013	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -56,6 +56,7 @@
 #include "pool_ip.h"
 #include "md5.h"
 #include "pool_stream.h"
+#include "pool_passwd.h"
 
 static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *timeout);
 static StartupPacket *read_startup_packet(POOL_CONNECTION *cp);
@@ -89,6 +90,9 @@ char remote_ps_data[NI_MAXHOST];		/* used for set_ps_display */
 
 volatile sig_atomic_t got_sighup = 0;
 
+char remote_host[NI_MAXHOST];	/* client host */
+char remote_port[NI_MAXSERV];	/* client port */
+
 /*
 * child main loop
 */
@@ -137,7 +141,11 @@ void do_child(int unix_fd, int inet_fd)
 
 	/* initialize random seed */
 	gettimeofday(&now, &tz);
+#if defined(sun) || defined(__sun)
+	srand((unsigned int) now.tv_usec);
+#else
 	srandom((unsigned int) now.tv_usec);
+#endif
 
 	/* initialize system db connection */
 	init_system_db_connection();
@@ -148,6 +156,15 @@ void do_child(int unix_fd, int inet_fd)
 		child_exit(1);
 	}
 
+	/*
+	 * Open pool_passwd in child process.  This is necessary to avoid the
+	 * file descriptor race condition reported in [pgpool-general: 1141].
+	 */
+	if (strcmp("", pool_config->pool_passwd))
+	{
+		pool_reopen_passwd_file();
+	}
+
 	timeout.tv_sec = pool_config->child_life_time;
 	timeout.tv_usec = 0;
 
@@ -479,9 +496,6 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	struct timeval *timeoutval;
 	struct timeval tv1, tv2, tmback = {0, 0};
 
-	char remote_host[NI_MAXHOST];
-	char remote_port[NI_MAXSERV];
-
 	set_ps_display("wait for connection request", false);
 
 	/* Destroy session context for just in case... */
@@ -634,7 +648,11 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd, struct timeval *time
 	{
 		pool_get_config(get_config_file_name(), RELOAD_CONFIG);
 		if (pool_config->enable_pool_hba)
+		{
 			load_hba(get_hba_file_name());
+			if (strcmp("", pool_config->pool_passwd))
+				pool_reopen_passwd_file();
+		}
 		if (pool_config->parallel_mode)
 			pool_memset_system_db_info(system_db_info->info);
 		got_sighup = 0;
@@ -816,19 +834,21 @@ static StartupPacket *read_startup_packet(POOL_CONNECTION *cp)
 	/* read startup packet length */
 	if (pool_read(cp, &len, sizeof(len)))
 	{
+		pool_error("read_startup_packet: incorrect packet length (%d)", len);
+		pool_free_startup_packet(sp);
+		alarm(0);
+		pool_signal(SIGALRM, SIG_IGN);
 		return NULL;
 	}
 	len = ntohl(len);
 	len -= sizeof(len);
 
-	if (len <= 0)
+	if (len <= 0 || len >= MAX_STARTUP_PACKET_LENGTH)
 	{
 		pool_error("read_startup_packet: incorrect packet length (%d)", len);
-	}
-	else if (len >= MAX_STARTUP_PACKET_LENGTH)
-	{
-		pool_error("read_startup_packet: invalid startup packet");
 		pool_free_startup_packet(sp);
+		alarm(0);
+		pool_signal(SIGALRM, SIG_IGN);
 		return NULL;
 	}
 
@@ -1383,6 +1403,12 @@ void pool_free_startup_packet(StartupPacket *sp)
  */
 void child_exit(int code)
 {
+	if (getpid() == mypid)
+	{
+		pool_log("child_exit: called from pgpool main. ignored.");
+		return;
+	}
+
 	/* count down global connection counter */
 	if (accepted)
 		connection_count_down();
@@ -1396,8 +1422,14 @@ void child_exit(int code)
 			pool_close(pool_system_db_connection()->con);
 	}
 
+	if (pool_config->memory_cache_enabled && !pool_is_shmem_cache())
+	{
+		memcached_disconnect();
+	}
+
 	/* let backend know now we are exiting */
-	send_frontend_exits();
+	if (pool_connection_pool)
+		send_frontend_exits();
 
 	exit(code);
 }
@@ -1406,7 +1438,7 @@ void child_exit(int code)
  * create a persistent connection
  */
 POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
-	char *hostname, int port, char *dbname, char *user, char *password)
+	char *hostname, int port, char *dbname, char *user, char *password, bool retry)
 {
 	POOL_CONNECTION_POOL_SLOT *cp;
 	int fd;
@@ -1445,11 +1477,11 @@ POOL_CONNECTION_POOL_SLOT *make_persistent_db_connection(
 	 */
 	if (*hostname == '/')
 	{
-		fd = connect_unix_domain_socket_by_port(port, hostname, TRUE);
+		fd = connect_unix_domain_socket_by_port(port, hostname, retry);
 	}
 	else
 	{
-		fd = connect_inet_domain_socket_by_port(hostname, port, TRUE);
+		fd = connect_inet_domain_socket_by_port(hostname, port, retry);
 	}
 
 	if (fd < 0)
@@ -1573,7 +1605,8 @@ void discard_persistent_db_connection(POOL_CONNECTION_POOL_SLOT *cp)
 }
 
 /*
- * do authentication for cp
+ * Do authentication. Assuming the only caller is
+ * *make_persistent_db_connection().
  */
 static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 {
@@ -1584,6 +1617,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	char state;
 	char *p;
 	int pid, key;
+	bool keydata_done;
 
 	/*
 	 * read kind expecting 'R' packet (authentication response)
@@ -1715,8 +1749,12 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 	}
 
 	/*
-	 * read pid etc.
+	 * Read backend key data and wait until Reay for query arriving or
+	 * error happens.
 	 */
+
+	keydata_done = false;
+
 	for (;;)
 	{
 		status = pool_read(cp->con, &kind, sizeof(kind));
@@ -1729,6 +1767,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 		switch (kind)
 		{
 			case 'K':	/* backend key data */
+				keydata_done = true;
 				pool_debug("s_do_auth: backend key data received");
 
 				/* read message length */
@@ -1758,21 +1797,9 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 					return -1;
 				}
 				cp->key = key;
+				break;
 
-				/* read kind expecting 'Z' (ready for query) */
-				status = pool_read(cp->con, &kind, sizeof(kind));
-				if (status < 0)
-				{
-					pool_error("s_do_auth: error while reading kind");
-					return -1;
-				}
-
-				if (kind != 'Z')
-				{
-					pool_error("s_do_auth: expecting Z got %c", kind);
-					return -1;
-				}
-
+			case 'Z':	/* Ready for query */
 				/* read message length */
 				status = pool_read(cp->con, &length, sizeof(length));
 				if (status < 0)
@@ -1792,16 +1819,22 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 
 				pool_debug("s_do_auth: transaction state: %c", state);
 				cp->con->tstate = state;
+
+				if (!keydata_done)
+				{
+					pool_error("s_do_auth: ready for query arrived before receiving keydata");
+				}
 				return 0;
 				break;
 
 			case 'S':	/* parameter status */
-				pool_debug("s_do_auth: parameter status data received");
-
+			case 'N':	/* notice response */
+			case 'E':	/* error response */
+				/* Just throw away data */
 				status = pool_read(cp->con, &length, sizeof(length));
 				if (status < 0)
 				{
-					pool_error("s_do_auth: error while reading message length");
+					pool_error("s_do_auth: error while reading message length. kind:%c", kind);
 					return -1;
 				}
 
@@ -1814,7 +1847,7 @@ static int s_do_auth(POOL_CONNECTION_POOL_SLOT *cp, char *password)
 				break;
 
 			default:
-				pool_error("s_do_auth: unknown response \"%c\" before processing BackendKeyData",
+				pool_error("s_do_auth: unknown response \"%c\" while processing BackendKeyData",
 						   kind);
 				break;
 		}
@@ -1898,7 +1931,13 @@ int select_load_balancing_node(void)
 			total_weight += BACKEND_INFO(i).backend_weight;
 		}
 	}
+
+#if defined(sun) || defined(__sun)
+	r = (((double)rand())/RAND_MAX) * total_weight;
+#else
 	r = (((double)random())/RAND_MAX) * total_weight;
+#endif
+
 	total_weight = 0.0;
 	for (i=0;i<NUM_BACKENDS;i++)
 	{
@@ -1958,7 +1997,7 @@ static void init_system_db_connection(void)
 																   pool_config->system_db_port,
 																   pool_config->system_db_dbname,
 																   pool_config->system_db_user,
-																   pool_config->system_db_password);
+																   pool_config->system_db_password, false);
 		if (system_db_info->connection == NULL)
 		{
 			pool_error("Could not make persistent system DB connection");
